#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
JK with analytic Fourier transformation
'''


import ctypes
import numpy
import numpy as np
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.lib import zdotNN, zdotCN, zdotNC
from pyscf.pbc import gto as pbcgto
from pyscf.gto.ft_ao import ft_aopair
from pyscf.pbc.df import ft_ao
from pyscf.pbc.df.df_jk import (_format_dms, _format_kpts_band, _format_jks,
                                _ewald_exxdiv_for_G0)
from pyscf.pbc.lib.kpts_helper import (is_zero, group_by_conj_pairs,
                                       kk_adapted_iter)
from pyscf.pbc.tools import k2gamma
from pyscf.pbc.df.incore import libpbc

def get_j_kpts(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None):
    if kpts_band is not None:
        return get_j_for_bands(mydf, dm_kpts, hermi, kpts, kpts_band)

    dm_kpts = lib.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    n_dm, nkpts, nao = dms.shape[:3]
    vj_kpts = numpy.zeros((n_dm,nkpts,nao,nao), dtype=numpy.complex128)
    kpt_allow = numpy.zeros(3)
    mesh = mydf.mesh
    coulG = mydf.weighted_coulG(kpt_allow, False, mesh)
    max_memory = (mydf.max_memory - lib.current_memory()[0]) * .8
    weight = 1./len(kpts)
    for Gpq, p0, p1 in mydf.ft_loop(mesh, kpt_allow, kpts, max_memory=max_memory,
                                    return_complex=False):
        _update_vj_(vj_kpts, Gpq, dms, coulG[p0:p1], weight)
        Gpq = None

    if is_zero(kpts):
        vj_kpts = vj_kpts.real.copy()
    return _format_jks(vj_kpts, dm_kpts, kpts_band, kpts)

def _update_vj_(vj_kpts, Gpq, dms, coulG, weight=None):
    r'''Compute the Coulomb matrix
    J_{kl} = \sum_{ij} \sum_G 4\pi/G^2 * FT(\rho_{ij}) IFT(\rho_{kl}) dm_{ji}
    for analytical FT tensor FT(\rho_{ij})
    '''
    # FT(\rho_{ij, kpt}) = \int exp(-i G*r) conj(\psi_{i,kpt}) \psi_{j,kpt} dr
    # IFT(\rho_{ij, kpt}) = \int exp(i G*r) conj(\psi_{i,kpt}) \psi_{j,kpt} dr
    # IFT(\rho_{ij, kpt}) = conj(FT(\rho_{ji, kpt}))
    # rhoG[k] = einsum('ij,gji->g', dms[i,k], IFT(\rho))
    #         = einsum('ij,gji->g', dms[i,k], aoao.conj().transpose(0,2,1))
    #         = einsum('ij,gij->g', dms[i,k].conj(), aoao).conj()
    # vG = sum_k rhoG[k] * coulG
    # vj[k] = einsum('g,gji->g', vG, aoao[k])
    is_real = vj_kpts.dtype == np.double
    GpqR, GpqI = Gpq
    rhoR = np.einsum('nkij,kgij->ng', dms.real, GpqR)
    rhoI = -np.einsum('nkij,kgij->ng', dms.real, GpqI)
    if not is_real:
        rhoR += np.einsum('nkij,kgij->ng', dms.imag, GpqI)
        rhoI += np.einsum('nkij,kgij->ng', dms.imag, GpqR)

    if weight is not None:
        coulG = coulG * weight
    vR = coulG * rhoR
    vI = coulG * rhoI

    vj_kpts.real += np.einsum('ng,kgij->nkij', vR, GpqR)
    vj_kpts.real -= np.einsum('ng,kgij->nkij', vI, GpqI)
    if not is_real:
        vj_kpts.imag += np.einsum('ng,kgij->nkij', vR, GpqI)
        vj_kpts.imag += np.einsum('ng,kgij->nkij', vI, GpqR)
    return vj_kpts

def get_j_for_bands(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None):
    log = logger.Logger(mydf.stdout, mydf.verbose)
    t1 = (logger.process_clock(), logger.perf_counter())

    dm_kpts = lib.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]

    kpt_allow = numpy.zeros(3)
    mesh = mydf.mesh
    coulG = mydf.weighted_coulG(kpt_allow, False, mesh)
    ngrids = len(coulG)
    rhoG = numpy.zeros((nset,ngrids), dtype=numpy.complex128)
    max_memory = (mydf.max_memory - lib.current_memory()[0]) * .8

    for aoaoks, p0, p1 in mydf.ft_loop(mesh, kpt_allow, kpts, max_memory=max_memory):
        for k, aoao in enumerate(aoaoks):
            rhoG[:,p0:p1] += numpy.einsum('nij,Lij->nL', dms[:,k].conj(),
                                          aoao.reshape(-1,nao,nao)).conj()
    aoao = aoaoks = p0 = p1 = None
    weight = 1./len(kpts)
    vG = rhoG * coulG * weight
    t1 = log.timer_debug1('get_j pass 1 to compute J(G)', *t1)

    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    nband = len(kpts_band)
    vj_kpts = numpy.zeros((nset,nband,nao,nao), dtype=numpy.complex128)
    for aoaoks, p0, p1 in mydf.ft_loop(mesh, kpt_allow, kpts_band,
                                        max_memory=max_memory):
        for k, aoao in enumerate(aoaoks):
            vj_kpts[:,k] += numpy.einsum('nL,Lij->nij', vG[:,p0:p1],
                                         aoao.reshape(-1,nao,nao))
    aoao = aoaoks = p0 = p1 = None

    if is_zero(kpts_band):
        vj_kpts = vj_kpts.real.copy()
    t1 = log.timer_debug1('get_j pass 2', *t1)
    return _format_jks(vj_kpts, dm_kpts, input_band, kpts)

def get_k_kpts(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None,
               exxdiv=None):
    if kpts_band is not None:
        return get_k_for_bands(mydf, dm_kpts, hermi, kpts, kpts_band, exxdiv)

    cpu0 = cpu1 = logger.process_clock(), logger.perf_counter()
    log = logger.new_logger(mydf)
    cell = mydf.cell
    mesh = mydf.mesh
    ngrids = np.prod(mesh)
    mo_coeff = getattr(dm_kpts, 'mo_coeff', None)
    mo_occ = getattr(dm_kpts, 'mo_occ', None)
    dm_kpts = np.asarray(dm_kpts)

    dms = _format_dms(dm_kpts, kpts)
    n_dm, nkpts, nao = dms.shape[:3]
    vkR = np.zeros((n_dm,nkpts,nao,nao))
    vkI = np.zeros((n_dm,nkpts,nao,nao))
    vk = [vkR, vkI]
    weight = 1. / nkpts

    aosym = 's1'
    bvk_kmesh = k2gamma.kpts_to_kmesh(cell, kpts)
    rcut = ft_ao.estimate_rcut(cell)
    supmol = ft_ao.ExtendedMole.from_cell(cell, bvk_kmesh, rcut.max())
    supmol = supmol.strip_basis(rcut)

    t_rev_pairs = group_by_conj_pairs(cell, kpts, return_kpts_pairs=False)
    try:
        t_rev_pairs = np.asarray(t_rev_pairs, dtype=np.int32, order='F')
    except TypeError:
        t_rev_pairs = [[k, k] if k_conj is None else [k, k_conj]
                       for k, k_conj in t_rev_pairs]
        t_rev_pairs = np.asarray(t_rev_pairs, dtype=np.int32, order='F')
    log.debug1('Num time-reversal pairs %d', len(t_rev_pairs))

    time_reversal_symmetry = mydf.time_reversal_symmetry
    if time_reversal_symmetry:
        for k, k_conj in t_rev_pairs:
            if k != k_conj and abs(dms[:,k_conj] - dms[:,k].conj()).max() > 1e-6:
                time_reversal_symmetry = False
                log.debug2('Disable time_reversal_symmetry')
                break

    if time_reversal_symmetry:
        k_to_compute = np.zeros(nkpts, dtype=np.int8)
        k_to_compute[t_rev_pairs[:,0]] = 1
    else:
        k_to_compute = np.ones(nkpts, dtype=np.int8)

    contract_mo_early = False
    if mo_coeff is None:
        dmsR = np.asarray(dms.real, order='C')
        dmsI = np.asarray(dms.imag, order='C')
        dm = [dmsR, dmsI]
        dm_factor = None

        if np.count_nonzero(k_to_compute) >= 2 * lib.num_threads():
            update_vk = _update_vk1_
        else:
            update_vk = _update_vk_
        log.debug2('set update_vk to %s', update_vk)
    else:
        # dm ~= dm_factor * dm_factor.T
        n_dm, nkpts, nao = dms.shape[:3]
        # mo_coeff, mo_occ are not a list of aligned array if
        # remove_lin_dep was applied to scf object
        if dm_kpts.ndim == 4:  # KUHF
            nocc = max(max(np.count_nonzero(x > 0) for x in z) for z in mo_occ)
            dm_factor = [[x[:,:nocc] for x in mo] for mo in mo_coeff]
            occs = [[x[:nocc] for x in z] for z in mo_occ]
        else:  # KRHF
            nocc = max(np.count_nonzero(x > 0) for x in mo_occ)
            dm_factor = [[mo[:,:nocc] for mo in mo_coeff]]
            occs = [[x[:nocc] for x in mo_occ]]
        dm_factor = np.array(dm_factor, dtype=np.complex128, order='C')
        dm_factor *= np.sqrt(np.array(occs, dtype=np.double))[:,:,None]

        bvk_ncells, rs_nbas, nimgs = supmol.bas_mask.shape
        s_nao = supmol.nao
        contract_mo_early = (time_reversal_symmetry and
                             bvk_ncells*nao*4 > s_nao*nocc*n_dm)
        log.debug2('time_reversal_symmetry = %s bvk_ncells = %d '
                   's_nao = %d nocc = %d n_dm = %d',
                   time_reversal_symmetry, bvk_ncells, s_nao, nocc, n_dm)
        log.debug2('Use algorithm contract_mo_early = %s', contract_mo_early)
        if contract_mo_early:
            s_nao = supmol.nao
            moR, moI = _mo_k2gamma(supmol, dm_factor, kpts, t_rev_pairs)
            if abs(moI).max() < 1e-5:
                dm = [moR, None]
                dm_factor = moR = moI = None
                ft_kern = _gen_ft_kernel_fake_gamma(cell, supmol, aosym)
                update_vk = _update_vk_fake_gamma
            else:
                moR = moI = None
                contract_mo_early = False

        if not contract_mo_early:
            dm = [np.asarray(dm_factor.real, order='C'),
                  np.asarray(dm_factor.imag, order='C')]
            dm_factor = None
            if np.count_nonzero(k_to_compute) >= 2 * lib.num_threads():
                update_vk = _update_vk1_dmf
            else:
                update_vk = _update_vk_dmf
        log.debug2('set update_vk to %s with dm_factor', update_vk)

    if not contract_mo_early:
        ft_kern = supmol.gen_ft_kernel(aosym, return_complex=False,
                                       kpts=kpts, verbose=log)

    Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
    Gv = np.asarray(Gv, order='F')
    gxyz = lib.cartesian_prod([np.arange(len(x)) for x in Gvbase])

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, (mydf.max_memory - mem_now))
    log.debug1('max_memory = %d MB (%d in use)', max_memory+mem_now, mem_now)

    if contract_mo_early:
        Gblksize = max(24, int((max_memory*1e6/16-nkpts*nao**2*3)/
                               (nao*s_nao+nao*nkpts*nocc))//8*8)
        Gblksize = min(Gblksize, ngrids, 200000)
        log.debug1('Gblksize = %d', Gblksize)
        buf = np.empty(Gblksize*s_nao*nao*2)
    else:
        Gblksize = max(24, int(max_memory*1e6/16/nao**2/(nkpts+3))//8*8)
        Gblksize = min(Gblksize, ngrids, 200000)
        log.debug1('Gblksize = %d', Gblksize)
        buf = np.empty(nkpts*Gblksize*nao**2*2)

    for group_id, (kpt, ki_idx, kj_idx, self_conj) \
            in enumerate(kk_adapted_iter(cell, kpts)):
        vkcoulG = mydf.weighted_coulG(kpt, exxdiv, mesh)
        for p0, p1 in lib.prange(0, ngrids, Gblksize):
            log.debug3('update_vk [%s:%s]', p0, p1)
            Gpq = ft_kern(Gv[p0:p1], gxyz[p0:p1], Gvbase, kpt, out=buf)
            update_vk(vk, Gpq, dm, vkcoulG[p0:p1] * weight, ki_idx, kj_idx,
                      not self_conj, k_to_compute, t_rev_pairs)
            Gpq = None
        cpu1 = log.timer_debug1(f'get_k_kpts group {group_id}', *cpu1)

    if is_zero(kpts) and not numpy.iscomplexobj(dm_kpts):
        vk_kpts = vkR
    else:
        vk_kpts = vkR + vkI * 1j

    # Add ewald_exxdiv contribution because G=0 was not included in the
    # non-uniform grids
    if exxdiv == 'ewald' and cell.low_dim_ft_type == 'inf_vacuum':
        _ewald_exxdiv_for_G0(cell, kpts, dms, vk_kpts, kpts)

    if time_reversal_symmetry:
        for k, k_conj in t_rev_pairs:
            if k != k_conj:
                vk_kpts[:,k_conj] = vk_kpts[:,k].conj()
    log.timer_debug1('get_k_kpts', *cpu0)
    return vk_kpts.reshape(dm_kpts.shape)

def get_k_for_bands(mydf, dm_kpts, hermi=1, kpts=numpy.zeros((1,3)), kpts_band=None,
                    exxdiv=None):
    cell = mydf.cell
    log = logger.Logger(mydf.stdout, mydf.verbose)
    t1 = (logger.process_clock(), logger.perf_counter())

    mesh = mydf.mesh
    dm_kpts = lib.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]

    swap_2e = (kpts_band is None)
    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    nband = len(kpts_band)
    kk_table = kpts_band.reshape(-1,1,3) - kpts.reshape(1,-1,3)
    kk_todo = numpy.ones(kk_table.shape[:2], dtype=bool)
    k_to_compute = np.ones(nkpts+len(kpts_band), dtype=np.int8)
    vkR = numpy.zeros((nset,nband,nao,nao))
    vkI = numpy.zeros((nset,nband,nao,nao))
    dmsR = numpy.asarray(dms.real, order='C')
    dmsI = numpy.asarray(dms.imag, order='C')

    mem_now = lib.current_memory()[0]
    max_memory = max(2000, (mydf.max_memory - mem_now)) * .8
    log.debug1('max_memory = %d MB (%d in use)', max_memory, mem_now)
    # K_pq = ( p{k1} i{k2} | i{k2} q{k1} )
    def make_kpt(kpt):  # kpt = kptj - kpti
        # search for all possible ki and kj that has ki-kj+kpt=0
        kk_match = numpy.einsum('ijx->ij', abs(kk_table + kpt)) < 1e-9
        kpti_idx, kptj_idx = numpy.where(kk_todo & kk_match)
        nkptj = len(kptj_idx)
        log.debug1('kpt = %s', kpt)
        log.debug2('kpti_idx = %s', kpti_idx)
        log.debug2('kptj_idx = %s', kptj_idx)
        kk_todo[kpti_idx,kptj_idx] = False
        if swap_2e and not is_zero(kpt):
            kk_todo[kptj_idx,kpti_idx] = False

        max_memory1 = max_memory * (nkptj+1)/(nkptj+5)
        #blksize = max(int(max_memory1*4e6/(nkptj+5)/16/nao**2), 16)

        #bufR = numpy.empty((blksize*nao**2))
        #bufI = numpy.empty((blksize*nao**2))
        # Use DF object to mimic KRHF/KUHF object in function get_coulG
        vkcoulG = mydf.weighted_coulG(kpt, exxdiv, mesh)
        weight = 1./len(kpts)
        perm_sym = swap_2e and not is_zero(kpt)
        for Gpq, p0, p1 in mydf.ft_loop(mesh, kpt, kpts, max_memory=max_memory1,
                                        return_complex=False):
            _update_vk_((vkR, vkI), Gpq, (dmsR, dmsI), vkcoulG[p0:p1]*weight,
                        kpti_idx, kptj_idx, perm_sym, k_to_compute, None)

    for ki, kpti in enumerate(kpts_band):
        for kj, kptj in enumerate(kpts):
            if kk_todo[ki,kj]:
                make_kpt(kptj-kpti)
        t1 = log.timer_debug1('get_k_kpts: make_kpt (%d,*)'%ki, *t1)

    if (is_zero(kpts) and is_zero(kpts_band) and
        not numpy.iscomplexobj(dm_kpts)):
        vk_kpts = vkR
    else:
        vk_kpts = vkR + vkI * 1j

    # Add ewald_exxdiv contribution because G=0 was not included in the
    # non-uniform grids
    if exxdiv == 'ewald' and cell.low_dim_ft_type == 'inf_vacuum':
        _ewald_exxdiv_for_G0(cell, kpts, dms, vk_kpts, kpts_band)

    return _format_jks(vk_kpts, dm_kpts, input_band, kpts)

def _update_vk_(vk, Gpq, dms, wcoulG, kpti_idx, kptj_idx, swap_2e,
                k_to_compute, t_rev_pairs):
    '''
    contraction for exchange matrices:

    vk += np.einsum('ngij,njk,nglk,g->nil', Gpq, dm, Gpq.conj(), coulG)
    vk += np.einsum('ngij,nli,nglk,g->nkj', Gpq, dm, Gpq.conj(), coulG)
    '''
    vkR, vkI = vk
    GpqR, GpqI = Gpq
    dmsR, dmsI = dms
    nG = len(wcoulG)
    n_dm, nkpts, nao = vkR.shape[:3]
    bufR = np.empty((nG*nao**2))
    bufI = np.empty((nG*nao**2))
    buf1R = np.empty((nG*nao**2))
    buf1I = np.empty((nG*nao**2))
    iLkR = np.ndarray((nao,nG,nao), buffer=buf1R)
    iLkI = np.ndarray((nao,nG,nao), buffer=buf1I)

    for k, (ki, kj) in enumerate(zip(kpti_idx, kptj_idx)):
        # case 1: k_pq = (pi|iq)
        #:v4 = np.einsum('ijL,lkL->ijkl', pqk, pqk.conj())
        #:vk += np.einsum('ijkl,jk->il', v4, dm)
        pLqR = np.ndarray((nao,nG,nao), buffer=bufR)
        pLqI = np.ndarray((nao,nG,nao), buffer=bufI)
        pLqR[:] = GpqR[kj].transpose(1,0,2)
        pLqI[:] = GpqI[kj].transpose(1,0,2)
        if k_to_compute[ki]:
            for i in range(n_dm):
                zdotNN(pLqR.reshape(-1,nao), pLqI.reshape(-1,nao),
                       dmsR[i,kj], dmsI[i,kj], 1,
                       iLkR.reshape(-1,nao), iLkI.reshape(-1,nao))
                iLkR *= wcoulG.reshape(1,nG,1)
                iLkI *= wcoulG.reshape(1,nG,1)
                zdotNC(iLkR.reshape(nao,-1), iLkI.reshape(nao,-1),
                       pLqR.reshape(nao,-1).T, pLqI.reshape(nao,-1).T,
                       1, vkR[i,ki], vkI[i,ki], 1)

        # case 2: k_pq = (iq|pi)
        #:v4 = np.einsum('iLj,lLk->ijkl', pqk, pqk.conj())
        #:vk += np.einsum('ijkl,li->kj', v4, dm)
        # <r|-G+k_rs|s> = conj(<s|G-k_rs|r>) = conj(<s|G+k_sr|r>)
        if swap_2e and k_to_compute[kj]:
            for i in range(n_dm):
                zdotNN(dmsR[i,ki], dmsI[i,ki], pLqR.reshape(nao,-1),
                       pLqI.reshape(nao,-1), 1,
                       iLkR.reshape(nao,-1), iLkI.reshape(nao,-1))
                iLkR *= wcoulG.reshape(1,nG,1)
                iLkI *= wcoulG.reshape(1,nG,1)
                zdotCN(pLqR.reshape(-1,nao).T, pLqI.reshape(-1,nao).T,
                       iLkR.reshape(-1,nao), iLkI.reshape(-1,nao),
                       1, vkR[i,kj], vkI[i,kj], 1)

def _update_vk1_(vk, Gpq, dms, wcoulG, kpti_idx, kptj_idx, swap_2e,
                 k_to_compute, t_rev_pairs):
    '''
    contraction for exchange matrices:

    vk += np.einsum('ngij,njk,nglk,g->nil', Gpq, dm, Gpq.conj(), coulG)
    vk += np.einsum('ngij,nli,nglk,g->nkj', Gpq, dm, Gpq.conj(), coulG)
    '''
    vkR, vkI = vk
    GpqR, GpqI = Gpq
    dmsR, dmsI = dms
    nG = len(wcoulG)
    n_dm, nkpts, nao = vkR.shape[:3]
    nkptj = len(kptj_idx)

    assert GpqR.transpose(0,2,3,1).flags.c_contiguous
    assert vkR.flags.c_contiguous
    assert dmsR.flags.c_contiguous
    assert kpti_idx.dtype == np.int32
    assert kptj_idx.dtype == np.int32
    assert k_to_compute.dtype == np.int8
    assert k_to_compute.size == nkpts

    libpbc.PBC_kcontract(
        vkR.ctypes.data_as(ctypes.c_void_p),
        vkI.ctypes.data_as(ctypes.c_void_p),
        dmsR.ctypes.data_as(ctypes.c_void_p),
        dmsI.ctypes.data_as(ctypes.c_void_p),
        GpqR.ctypes.data_as(ctypes.c_void_p),
        GpqI.ctypes.data_as(ctypes.c_void_p),
        wcoulG.ctypes.data_as(ctypes.c_void_p),
        kpti_idx.ctypes.data_as(ctypes.c_void_p),
        kptj_idx.ctypes.data_as(ctypes.c_void_p),
        k_to_compute.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(swap_2e), ctypes.c_int(n_dm), ctypes.c_int(nao),
        ctypes.c_int(nG), ctypes.c_int(nkpts), ctypes.c_int(nkptj))

def _update_vk_dmf(vk, Gpq, dmf, wcoulG, kpti_idx, kptj_idx, swap_2e,
                   k_to_compute, t_rev_pairs):
    '''
    dmf is the factorized dm, dm = dmf * dmf.conj().T
    Computing exchange matrices with dmf:
    vk += np.einsum('ngij,njp,nkp,nglk,g->nil', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    vk += np.einsum('ngij,nlp,nip,nglk,g->nkj', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    '''
    vkR, vkI = vk
    GpqR, GpqI = Gpq
    dmfR, dmfI = dmf
    nG = len(wcoulG)
    n_dm, nkpts, nao, nocc = dmfR.shape
    bufR = np.empty((nG*nao**2))
    bufI = np.empty((nG*nao**2))
    bufR1 = np.empty((nG*nao*nocc))
    bufI1 = np.empty((nG*nao*nocc))

    for k, (ki, kj) in enumerate(zip(kpti_idx, kptj_idx)):
        # case 1: k_pq = (pi|iq)
        #:v4 = np.einsum('ijL,lkL->ijkl', pqk, pqk.conj())
        #:vk += np.einsum('ijkl,jk->il', v4, dm)
        pLqR = np.ndarray((nao,nG,nao), buffer=bufR)
        pLqI = np.ndarray((nao,nG,nao), buffer=bufI)
        pLqR[:] = GpqR[kj].transpose(1,0,2)
        pLqI[:] = GpqI[kj].transpose(1,0,2)
        if k_to_compute[ki]:
            iLkR = np.ndarray((nao,nG,nocc), buffer=bufR1)
            iLkI = np.ndarray((nao,nG,nocc), buffer=bufI1)
            for i in range(n_dm):
                zdotNN(pLqR.reshape(-1,nao), pLqI.reshape(-1,nao),
                       dmfR[i,kj], dmfI[i,kj], 1,
                       iLkR.reshape(-1,nocc), iLkI.reshape(-1,nocc))
                iLkR1 = iLkR * wcoulG[:,None]
                iLkI1 = iLkI * wcoulG[:,None]
                zdotNC(iLkR1.reshape(nao,-1), iLkI1.reshape(nao,-1),
                       iLkR.reshape(nao,-1).T, iLkI.reshape(nao,-1).T,
                       1, vkR[i,ki], vkI[i,ki], 1)

        # case 2: k_pq = (iq|pi)
        #:v4 = np.einsum('iLj,lLk->ijkl', pqk, pqk.conj())
        #:vk += np.einsum('ijkl,li->kj', v4, dm)
        # <r|-G+k_rs|s> = conj(<s|G-k_rs|r>) = conj(<s|G+k_sr|r>)
        if swap_2e and k_to_compute[kj]:
            iLkR = np.ndarray((nocc,nG,nao), buffer=bufR1)
            iLkI = np.ndarray((nocc,nG,nao), buffer=bufI1)
            for i in range(n_dm):
                zdotCN(dmfR[i,ki].T, dmfI[i,ki].T, pLqR.reshape(nao,-1),
                       pLqI.reshape(nao,-1), 1,
                       iLkR.reshape(nocc,-1), iLkI.reshape(nocc,-1))
                iLkR1 = iLkR * wcoulG[:,None]
                iLkI1 = iLkI * wcoulG[:,None]
                zdotCN(iLkR.reshape(-1,nao).T, iLkI.reshape(-1,nao).T,
                       iLkR1.reshape(-1,nao), iLkI1.reshape(-1,nao),
                       1, vkR[i,kj], vkI[i,kj], 1)

def _update_vk1_dmf(vk, Gpq, dmf, wcoulG, kpti_idx, kptj_idx, swap_2e,
                    k_to_compute, t_rev_pairs):
    '''
    dmf is the factorized dm, dm = dmf * dmf.conj().T
    Computing exchange matrices with dmf:
    vk += np.einsum('ngij,njp,nkp,nglk,g->nil', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    vk += np.einsum('ngij,nlp,nip,nglk,g->nkj', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    '''
    vkR, vkI = vk
    GpqR, GpqI = Gpq
    dmfR, dmfI = dmf
    nG = len(wcoulG)
    n_dm, nkpts, nao, nocc = dmfR.shape
    nkptj = len(kptj_idx)

    assert GpqR.transpose(0,2,3,1).flags.c_contiguous
    assert vkR.flags.c_contiguous
    assert dmfI.flags.c_contiguous
    assert kpti_idx.dtype == np.int32
    assert kptj_idx.dtype == np.int32
    assert k_to_compute.dtype == np.int8
    assert k_to_compute.size == nkpts

    libpbc.PBC_kcontract_dmf(
        vkR.ctypes.data_as(ctypes.c_void_p),
        vkI.ctypes.data_as(ctypes.c_void_p),
        dmfR.ctypes.data_as(ctypes.c_void_p),
        dmfI.ctypes.data_as(ctypes.c_void_p),
        GpqR.ctypes.data_as(ctypes.c_void_p),
        GpqI.ctypes.data_as(ctypes.c_void_p),
        wcoulG.ctypes.data_as(ctypes.c_void_p),
        kpti_idx.ctypes.data_as(ctypes.c_void_p),
        kptj_idx.ctypes.data_as(ctypes.c_void_p),
        k_to_compute.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(swap_2e),
        ctypes.c_int(n_dm), ctypes.c_int(nao), ctypes.c_int(nocc),
        ctypes.c_int(nG), ctypes.c_int(nkpts), ctypes.c_int(nkptj))

def _update_vk_fake_gamma(vk, Gpq, dmf, wcoulG, kpti_idx, kptj_idx, swap_2e,
                          k_to_compute, t_rev_pairs):
    '''
    dmf is the factorized dm, dm = dmf * dmf.conj().T
    Computing exchange matrices with dmf:
    vk += np.einsum('ngij,njp,nkp,nglk,g->nil', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    vk += np.einsum('ngij,nlp,nip,nglk,g->nkj', Gpq, dmf, dmf.conj(), Gpq.conj(), coulG)
    '''
    vkR, vkI = vk
    GpqR, GpqI = Gpq
    dmfR, dmfI = dmf
    nG = len(wcoulG)
    n_dm, s_nao, nkpts, nocc = dmfR.shape
    nao = vkR.shape[-1]

    GpqR = GpqR.transpose(2,1,0)
    assert GpqR.flags.c_contiguous
    assert kpti_idx.dtype == np.int32
    assert kptj_idx.dtype == np.int32
    assert k_to_compute.dtype == np.int8
    assert k_to_compute.size == nkpts
    assert kptj_idx.size == nkpts

    for i in range(n_dm):
        libpbc.PBC_kcontract_fake_gamma(
            vkR[i].ctypes.data_as(ctypes.c_void_p),
            vkI[i].ctypes.data_as(ctypes.c_void_p),
            dmfR[i].ctypes.data_as(ctypes.c_void_p),
            GpqR.ctypes.data_as(ctypes.c_void_p),
            GpqI.ctypes.data_as(ctypes.c_void_p),
            wcoulG.ctypes.data_as(ctypes.c_void_p),
            kpti_idx.ctypes.data_as(ctypes.c_void_p),
            kptj_idx.ctypes.data_as(ctypes.c_void_p),
            k_to_compute.ctypes.data_as(ctypes.c_void_p),
            t_rev_pairs.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(t_rev_pairs.shape[0]),
            ctypes.c_int(swap_2e), ctypes.c_int(s_nao), ctypes.c_int(nao),
            ctypes.c_int(nocc), ctypes.c_int(nG), ctypes.c_int(nkpts))

def _mo_k2gamma(supmol, dm_factor, kpts, t_rev_pairs):
    '''Transform to MO of supcell at gamma point'''
    sh_loc = supmol.sh_loc  # maps the bvk shell Id to supmol shell Id
    ao_loc = supmol.ao_loc
    rs_cell = supmol.rs_cell
    cell0_ao_loc = rs_cell.ao_loc
    nbasp = rs_cell.ref_cell.nbas
    bvk_ncells, _, nimgs = supmol.bas_mask.shape
    expLk = np.exp(1j*np.dot(supmol.bvkmesh_Ls, kpts.T))
    expLk = np.asarray(expLk, order='C')
    s_nao = supmol.nao
    n_dm, nkpts, nao, nocc = dm_factor.shape
    moR = np.empty((n_dm,s_nao,nkpts,nocc))
    moI = np.empty((n_dm,s_nao,nkpts,nocc))
    t_rev_pairs = np.asarray(t_rev_pairs, dtype=np.int32, order='F')
    for i_dm in range(n_dm):
        libpbc.PBCmo_k2gamma(
            moR[i_dm].ctypes.data_as(ctypes.c_void_p),
            moI[i_dm].ctypes.data_as(ctypes.c_void_p),
            dm_factor[i_dm].ctypes.data_as(ctypes.c_void_p),
            expLk.ctypes.data_as(ctypes.c_void_p),
            sh_loc.ctypes.data_as(ctypes.c_void_p),
            ao_loc.ctypes.data_as(ctypes.c_void_p),
            cell0_ao_loc.ctypes.data_as(ctypes.c_void_p),
            t_rev_pairs.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(t_rev_pairs.shape[0]),
            ctypes.c_int(bvk_ncells), ctypes.c_int(nbasp),
            ctypes.c_int(s_nao), ctypes.c_int(nao), ctypes.c_int(nocc),
            ctypes.c_int(nkpts), ctypes.c_int(nimgs))
    return moR, moI

def _gen_ft_kernel_fake_gamma(cell, supmol, aosym='s1'):
    ovlp_mask = supmol.get_ovlp_mask()
    ovlp_mask = np.asarray(ovlp_mask, dtype=np.int8, order='F')
    cell = supmol.rs_cell
    supmol1 = cell.to_mol() + supmol
    shls_slice = (0, cell.nbas, cell.nbas, supmol1.nbas)
    b = cell.reciprocal_vectors()

    def ft_kern(Gv, gxyz=None, Gvbase=None, q=None, kpts=None, out=None):
        return ft_aopair(supmol1, Gv, shls_slice, aosym, b,
                         gxyz=gxyz, Gvbase=Gvbase, out=out,
                         q=q, return_complex=False, ovlp_mask=ovlp_mask)
    return ft_kern

##################################################
#
# Single k-point
#
##################################################

def get_jk(mydf, dm, hermi=1, kpt=numpy.zeros(3),
           kpts_band=None, with_j=True, with_k=True, exxdiv=None):
    '''JK for given k-point'''
    vj = vk = None
    if kpts_band is not None and abs(kpt-kpts_band).sum() > 1e-9:
        kpt = numpy.reshape(kpt, (1,3))
        if with_k:
            vk = get_k_kpts(mydf, dm, hermi, kpt, kpts_band, exxdiv)
        if with_j:
            vj = get_j_kpts(mydf, dm, hermi, kpt, kpts_band)
        return vj, vk

    cell = mydf.cell
    log = logger.Logger(mydf.stdout, mydf.verbose)
    t1 = (logger.process_clock(), logger.perf_counter())

    dm = numpy.asarray(dm, order='C')
    dms = _format_dms(dm, kpt.reshape(1, 3))
    nset, _, nao = dms.shape[:3]
    dms = dms.reshape(nset,nao,nao)
    j_real = is_zero(kpt)
    k_real = is_zero(kpt) and not numpy.iscomplexobj(dms)

    mesh = mydf.mesh
    kptii = numpy.asarray((kpt,kpt))
    kpt_allow = numpy.zeros(3)

    if with_j:
        vjcoulG = mydf.weighted_coulG(kpt_allow, False, mesh)
        vjR = numpy.zeros((nset,nao,nao))
        vjI = numpy.zeros((nset,nao,nao))
    if with_k:
        vkcoulG = mydf.weighted_coulG(kpt_allow, exxdiv, mesh)
        vkR = numpy.zeros((nset,nao,nao))
        vkI = numpy.zeros((nset,nao,nao))
    dmsR = numpy.asarray(dms.real.reshape(nset,nao,nao), order='C')
    dmsI = numpy.asarray(dms.imag.reshape(nset,nao,nao), order='C')
    mem_now = lib.current_memory()[0]
    max_memory = max(2000, (mydf.max_memory - mem_now)) * .8
    log.debug1('max_memory = %d MB (%d in use)', max_memory, mem_now)
    t2 = t1

    # rho_rs(-G+k_rs) is computed as conj(rho_{rs^*}(G-k_rs))
    #                 == conj(transpose(rho_sr(G+k_sr), (0,2,1)))
    #blksize = max(int(max_memory*.25e6/16/nao**2), 16)
    pLqR = pLqI = None
    for pqkR, pqkI, p0, p1 in mydf.pw_loop(mesh, kptii, max_memory=max_memory):
        t2 = log.timer_debug1('%d:%d ft_aopair'%(p0,p1), *t2)
        pqkR = pqkR.reshape(nao,nao,-1)
        pqkI = pqkI.reshape(nao,nao,-1)
        if with_j:
            #:v4 = numpy.einsum('ijL,lkL->ijkl', pqk, pqk.conj())
            #:vj += numpy.einsum('ijkl,lk->ij', v4, dm)
            for i in range(nset):
                rhoR = numpy.einsum('pq,pqk->k', dmsR[i], pqkR)
                rhoR+= numpy.einsum('pq,pqk->k', dmsI[i], pqkI)
                rhoI = numpy.einsum('pq,pqk->k', dmsI[i], pqkR)
                rhoI-= numpy.einsum('pq,pqk->k', dmsR[i], pqkI)
                rhoR *= vjcoulG[p0:p1]
                rhoI *= vjcoulG[p0:p1]
                vjR[i] += numpy.einsum('pqk,k->pq', pqkR, rhoR)
                vjR[i] -= numpy.einsum('pqk,k->pq', pqkI, rhoI)
                if not j_real:
                    vjI[i] += numpy.einsum('pqk,k->pq', pqkR, rhoI)
                    vjI[i] += numpy.einsum('pqk,k->pq', pqkI, rhoR)
        #t2 = log.timer_debug1('        with_j', *t2)

        if with_k:
            #:v4 = numpy.einsum('ijL,lkL->ijkl', pqk, pqk.conj())
            #:vk += numpy.einsum('ijkl,jk->il', v4, dm)
            pLqR = lib.transpose(pqkR, axes=(0,2,1), out=pLqR).reshape(-1,nao)
            pLqI = lib.transpose(pqkI, axes=(0,2,1), out=pLqI).reshape(-1,nao)
            nG = p1 - p0
            iLkR = numpy.ndarray((nao,nG,nao), buffer=pqkR)
            iLkI = numpy.ndarray((nao,nG,nao), buffer=pqkI)
            for i in range(nset):
                if k_real:
                    lib.dot(pLqR, dmsR[i], 1, iLkR.reshape(nao*nG,nao))
                    lib.dot(pLqI, dmsR[i], 1, iLkI.reshape(nao*nG,nao))
                    iLkR *= vkcoulG[p0:p1].reshape(1,nG,1)
                    iLkI *= vkcoulG[p0:p1].reshape(1,nG,1)
                    lib.dot(iLkR.reshape(nao,-1), pLqR.reshape(nao,-1).T, 1, vkR[i], 1)
                    lib.dot(iLkI.reshape(nao,-1), pLqI.reshape(nao,-1).T, 1, vkR[i], 1)
                else:
                    zdotNN(pLqR, pLqI, dmsR[i], dmsI[i], 1,
                           iLkR.reshape(-1,nao), iLkI.reshape(-1,nao))
                    iLkR *= vkcoulG[p0:p1].reshape(1,nG,1)
                    iLkI *= vkcoulG[p0:p1].reshape(1,nG,1)
                    zdotNC(iLkR.reshape(nao,-1), iLkI.reshape(nao,-1),
                           pLqR.reshape(nao,-1).T, pLqI.reshape(nao,-1).T,
                           1, vkR[i], vkI[i], 1)
            #t2 = log.timer_debug1('        with_k', *t2)
        pqkR = pqkI = pLqR = pLqI = iLkR = iLkI = None
        #t2 = log.timer_debug1('%d:%d'%(p0,p1), *t2)
    t1 = log.timer_debug1('aft_jk.get_jk', *t1)

    if with_j:
        if j_real:
            vj = vjR
        else:
            vj = vjR + vjI * 1j
        vj = vj.reshape(dm.shape)
    if with_k:
        if k_real:
            vk = vkR
        else:
            vk = vkR + vkI * 1j
        # Add ewald_exxdiv contribution because G=0 was not included in the
        # non-uniform grids
        if exxdiv == 'ewald' and cell.low_dim_ft_type == 'inf_vacuum':
            _ewald_exxdiv_for_G0(cell, kpt, dms, vk)
        vk = vk.reshape(dm.shape)
    return vj, vk



#!/usr/bin/env python
# Copyright 2018-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Pseudo-spectral methods (COSX, PS, SN-K)
'''

import numpy
from pyscf import lib
from pyscf import gto
from pyscf import scf
from pyscf import mcscf
from pyscf.scf import _vhf
from pyscf.lib import logger
from pyscf.sgx import sgx_jk
from pyscf.df import df_jk
from pyscf import __config__

def sgx_fit(mf, auxbasis=None, with_df=None, pjs=False):
    '''For the given SCF object, update the J, K matrix constructor with
    corresponding SGX or density fitting integrals.

    Args:
        mf : an SCF object

    Kwargs:
        auxbasis : str or basis dict
            Same format to the input attribute mol.basis.  If auxbasis is
            None, optimal auxiliary basis based on AO basis (if possible) or
            even-tempered Gaussian basis will be used.
        with_df : SGX
            Existing SGX object for the system.
        pjs: bool
            Whether to perform P-junction screening (screening matrix elements
            by the density matrix). Default False. If True, dfj is set to True
            automatically at the beginning of the calculation, as this screening
            is only for K-matrix elements.

    Returns:
        An SCF object with a modified J, K matrix constructor which uses density
        fitting integrals to compute J and K

    Examples:

    >>> mol = gto.M(atom='H 0 0 0; F 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = sgx_fit(scf.RHF(mol))
    >>> mf.scf()
    -100.00978770917165

    >>> mol.symmetry = 1
    >>> mol.build(0, 0)
    >>> mf = sgx_fit(scf.UHF(mol))
    >>> mf.scf()
    -100.00978770951018
    '''
    from pyscf.df.addons import predefined_auxbasis
    assert (isinstance(mf, scf.hf.SCF))

    if with_df is None:
        mol = mf.mol
        if auxbasis is None:
            if isinstance(mf, scf.hf.KohnShamDFT):
                xc = mf.xc
            else:
                xc = 'HF'
            if xc == 'LDA,VWN':
                # This is likely the default xc setting of a KS instance.
                # Postpone the auxbasis assignment to with_df.build().
                auxbasis = None
            else:
                auxbasis = predefined_auxbasis(mol, mol.basis, xc)
        with_df = SGX(mol, auxbasis=auxbasis, pjs=pjs)
        with_df.max_memory = mf.max_memory
        with_df.stdout = mf.stdout
        with_df.verbose = mf.verbose

    if isinstance(mf, _SGXHF):
        mf = mf.copy()
        mf.with_df = with_df
        return mf

    dfmf = _SGXHF(mf, with_df, auxbasis)
    return lib.set_class(dfmf, (_SGXHF, mf.__class__))

# A tag to label the derived SCF class
class _SGXHF:

    __name_mixin__ = 'SGX'

    _keys = {
        'auxbasis', 'with_df', 'direct_scf_sgx', 'rebuild_nsteps'
    }

    def __init__(self, mf, df=None, auxbasis=None):
        self.__dict__.update(mf.__dict__)
        self._eri = None
        self.auxbasis = auxbasis
        self.with_df = df

        # Grids/Integral quality varies during SCF. VHF cannot be
        # constructed incrementally through standard direct SCF.
        self.direct_scf = False
        # Set direct_scf_sgx True to use direct SCF for each
        # grid size with SGX.
        self.direct_scf_sgx = False
        # Set rebuild_nsteps to control how many direct SCF steps
        # are taken between resets of the SGX JK matrix.
        # Default 5, only used if direct_scf_sgx = True
        self.rebuild_nsteps = 5

        self._last_dm = 0
        self._last_vj = 0
        self._last_vk = 0
        self._in_scf = False

    def undo_sgx(self):
        obj = lib.view(self, lib.drop_class(self.__class__, _SGXHF))
        del obj.auxbasis
        del obj.with_df
        del obj.direct_scf_sgx
        del obj.rebuild_nsteps
        del obj._in_scf
        return obj

    def build(self, mol=None, **kwargs):
        if self.direct_scf_sgx:
            self._nsteps_direct = 0
            self._last_dm = 0
            self._last_vj = 0
            self._last_vk = 0
        if self.direct_scf:
            self.with_df.build(level=self.with_df.grids_level_f)
        else:
            self.with_df.build(level=self.with_df.grids_level_i)
        if self.with_df.pjs:
            if not self.with_df.dfj:
                import warnings
                msg = '''
                P-junction screening is not compatible with SGX J-matrix.
                Setting dfj = True. If you want to use SGX J-matrix,
                set pjs = False to turn off P-junction screening.
                '''
                warnings.warn(msg)
            self.with_df.dfj = True # no SGX-J allowed if P-junction screening on
        return super().build(mol, **kwargs)

    def reset(self, mol=None):
        self.with_df.reset(mol)
        return super().reset(mol)

    def pre_kernel(self, envs):
        self.direct_scf = False # should always be False
        if self.with_df.grids_level_i != self.with_df.grids_level_f:
            self._in_scf = True

    def get_jk(self, mol=None, dm=None, hermi=1, with_j=True, with_k=True,
               omega=None):
        if dm is None: dm = self.make_rdm1()
        with_df = self.with_df
        if not with_df:
            return super().get_jk(self, mol, dm, hermi, with_j, with_k, omega)
        if (self._opt.get(omega) is None and
            self.with_df.direct_j and (not self.with_df.dfj)):
            with mol.with_range_coulomb(omega):
                self._opt[omega] = self.init_direct_scf(mol)
        vhfopt = self._opt.get(omega)

        if self._in_scf and not self.direct_scf:
            if numpy.linalg.norm(dm - self._last_dm) < with_df.grids_switch_thrd \
                    and with_df.grids_level_f != with_df.grids_level_i:
                # only reset if grids_level_f and grids_level_i differ
                logger.debug(self, 'Switching SGX grids')
                with_df.build(level=with_df.grids_level_f)
                self._nsteps_direct = 0
                self._in_scf = False
                self._last_dm = 0
                self._last_vj = 0
                self._last_vk = 0

        if self.direct_scf_sgx:
            vj, vk = with_df.get_jk(dm-self._last_dm, hermi, vhfopt,
                                    with_j, with_k,
                                    self.direct_scf_tol, omega)
            vj += self._last_vj
            vk += self._last_vk
            self._last_dm = numpy.asarray(dm)
            self._last_vj = vj.copy()
            self._last_vk = vk.copy()
            self._nsteps_direct += 1
            if self.rebuild_nsteps > 0 and \
                    self._nsteps_direct >= self.rebuild_nsteps:
                logger.debug(self, 'Resetting JK matrix')
                self._nsteps_direct = 0
                self._last_dm = 0
                self._last_vj = 0
                self._last_vk = 0
        else:
            self._last_dm = numpy.asarray(dm)
            vj, vk = with_df.get_jk(dm, hermi, vhfopt, with_j, with_k,
                                    self.direct_scf_tol, omega)

        return vj, vk

    def post_kernel(self, envs):
        self._in_scf = False
        self._last_dm = 0
        self._last_vj = 0
        self._last_vk = 0

    def to_gpu(self):
        raise NotImplementedError

    def method_not_implemented(self, *args, **kwargs):
        raise NotImplementedError
    nuc_grad_method = Gradients = method_not_implemented
    Hessian = method_not_implemented
    NMR = method_not_implemented
    NSR = method_not_implemented
    Polarizability = method_not_implemented
    RotationalGTensor = method_not_implemented
    MP2 = method_not_implemented
    CISD = method_not_implemented
    CCSD = method_not_implemented
    CASCI = method_not_implemented
    CASSCF = method_not_implemented

scf.hf.SCF.COSX = sgx_fit
mcscf.casci.CASBase.COSX = sgx_fit


def _make_opt(mol, pjs=False,
              direct_scf_tol=getattr(__config__, 'scf_hf_SCF_direct_scf_tol', 1e-13)):
    '''Optimizer to generate 3-center 2-electron integrals'''
    if pjs:
        vhfopt = _vhf.SGXOpt(mol, 'int1e_grids', 'SGXnr_ovlp_prescreen',
                             dmcondname='SGXnr_dm_cond',
                             direct_scf_tol=direct_scf_tol)
    else:
        vhfopt = _vhf._VHFOpt(mol, 'int1e_grids', 'SGXnr_ovlp_prescreen',
                              direct_scf_tol=direct_scf_tol)
    vhfopt.init_cvhf_direct(mol, 'int1e_ovlp', 'SGXnr_q_cond')
    return vhfopt


class SGX(lib.StreamObject):
    _keys = {
        'mol', 'grids_thrd', 'grids_level_i', 'grids_level_f',
        'grids_switch_thrd', 'dfj', 'direct_j', 'pjs', 'debug', 'grids',
        'blockdim', 'auxmol',
    }

    def __init__(self, mol, auxbasis=None, pjs=False):
        self.mol = mol
        self.stdout = mol.stdout
        self.verbose = mol.verbose
        self.max_memory = mol.max_memory
        self.grids_thrd = 1e-10
        self.grids_level_i = 0  # initial grids level
        self.grids_level_f = 1  # final grids level
        self.grids_switch_thrd = 0.03
        # compute J matrix using DF and K matrix using SGX. It's identical to
        # the RIJCOSX method in ORCA
        self.dfj = False
        self.direct_j = False
        self._auxbasis = auxbasis
        self.pjs = pjs

        # debug=True generates a dense tensor of the Coulomb integrals at each
        # grids. debug=False utilizes the sparsity of the integral tensor and
        # contracts the sparse tensor and density matrices on the fly.
        self.debug = False

        self.grids = None
        self.blockdim = 1200
        self.auxmol = None
        self._vjopt = None
        self._opt = None
        self._last_dm = 0
        self._rsh_df = {}  # Range separated Coulomb DF objects

    @property
    def auxbasis(self):
        return self._auxbasis
    @auxbasis.setter
    def auxbasis(self, x):
        if self._auxbasis != x:
            self._auxbasis = x
            self.auxmol = None

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('******** %s ********', self.__class__)
        log.info('max_memory = %s', self.max_memory)
        log.info('grids_level_i = %s', self.grids_level_i)
        log.info('grids_level_f = %s', self.grids_level_f)
        log.info('grids_thrd = %s', self.grids_thrd)
        log.info('grids_switch_thrd = %s', self.grids_switch_thrd)
        log.info('dfj = %s', self.dfj)
        log.info('auxbasis = %s', self.auxbasis)
        return self

    # To mimic DF object, so that SGX can be used as in DF-SCF method by setting
    # mf.with_df = SGX(mol)
    @property
    def _cderi(self):
        return self.grids

    def build(self, level=None):
        if level is None:
            level = self.grids_level_f
        self.grids = sgx_jk.get_gridss(self.mol, level, self.grids_thrd)
        self._opt = _make_opt(self.mol, pjs=self.pjs)

        # In the RSH-integral temporary treatment, recursively rebuild SGX
        # objects in _rsh_df.
        if self._rsh_df:
            for k, v in self._rsh_df.items():
                v.build(level)
        return self

    def kernel(self, *args, **kwargs):
        return self.build(*args, **kwargs)

    def reset(self, mol=None):
        '''Reset mol and clean up relevant attributes for scanner mode'''
        if mol is not None:
            self.mol = mol
        self.grids = None
        self.auxmol = None
        self._vjopt = None
        self._opt = None
        self._last_dm = 0
        self._rsh_df = {}
        return self

    def get_jk(self, dm, hermi=1, vhfopt=None, with_j=True, with_k=True,
               direct_scf_tol=getattr(__config__, 'scf_hf_SCF_direct_scf_tol', 1e-13),
               omega=None):
        if omega is not None:
            # A temporary treatment for RSH integrals
            key = '%.6f' % omega
            if key in self._rsh_df:
                rsh_df = self._rsh_df[key]
            else:
                rsh_df = self.copy()
                rsh_df._rsh_df = None  # to avoid circular reference
                # Not all attributes need to be reset. Resetting _vjopt
                # because it is used by get_j method of regular DF object.
                rsh_df._vjopt = None
                self._rsh_df[key] = rsh_df
                logger.info(self, 'Create RSH-SGX object %s for omega=%s', rsh_df, omega)

            with rsh_df.mol.with_range_coulomb(omega):
                return rsh_df.get_jk(dm, hermi, with_j, with_k,
                                     direct_scf_tol)

        if with_j and self.dfj:
            vj = df_jk.get_j(self, dm, hermi, direct_scf_tol)
            if with_k:
                vk = sgx_jk.get_jk(self, dm, hermi, False, with_k, direct_scf_tol)[1]
            else:
                vk = None
        elif with_j and self.direct_j:
            vj, _ = _vhf.direct(dm, self.mol._atm, self.mol._bas, self.mol._env,
                                vhfopt, hermi, self.mol.cart, True, False)
            if with_k:
                vk = sgx_jk.get_jk(self, dm, hermi, False, with_k, direct_scf_tol)[1]
            else:
                vk = None
        else:
            vj, vk = sgx_jk.get_jk(self, dm, hermi, with_j, with_k, direct_scf_tol)
        return vj, vk

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2025 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
DFT+U on molecules

See also the pbc.dft.krkspu and pbc.dft.kukspu module
'''

import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.data.nist import HARTREE2EV
from pyscf.dft import uks
from pyscf.dft.rkspu import set_U, make_minao_lo

def get_veff(ks, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
    """
    Coulomb + XC functional + (Hubbard - double counting) for UKS+U.
    """
    if mol is None: mol = ks.mol
    if dm is None: dm = ks.make_rdm1()

    # J + V_xc
    vxc = uks.get_veff(ks, mol, dm, dm_last=dm_last, vhf_last=vhf_last,
                       hermi=hermi)

    # V_U
    C_ao_lo = ks.C_ao_lo
    ovlp = ks.get_ovlp()
    C_inv = np.dot(C_ao_lo.conj().T, ovlp)
    rdm1_lo = [C_inv.dot(dm[0]).dot(C_inv.conj().T),
               C_inv.dot(dm[1]).dot(C_inv.conj().T)]

    alphas = ks.alpha
    if not hasattr(alphas, '__len__'): # not a list or tuple
        alphas = [alphas] * len(ks.U_idx)

    E_U = 0.0
    logger.info(ks, "-" * 79)
    with np.printoptions(precision=5, suppress=True, linewidth=1000):
        for idx, val, lab, alpha in zip(ks.U_idx, ks.U_val, ks.U_lab, alphas):
            lab_string = " "
            for l in lab:
                lab_string += "%9s" %(l.split()[-1])
            lab_sp = lab[0].split()
            logger.info(ks, "local rdm1 of atom %s: ",
                        " ".join(lab_sp[:2]) + " " + lab_sp[2][:2])
            for s in range(2):
                P = rdm1_lo[s][idx[:,None], idx]
                SC = np.dot(ovlp, C_ao_lo[:, idx])
                loc_sites = P.shape[-1]
                vhub_loc = (np.eye(loc_sites) - P * 2.0) * (val * 0.5)
                if alpha is not None:
                    # LR-cDFT perturbation for Hubbard U
                    E_U += alpha * P.trace()
                    vhub_loc += np.eye(loc_sites) * alpha
                vxc[s] += SC.dot(vhub_loc).dot(SC.conj().T)
                E_U += (val * 0.5) * (P.trace() - np.dot(P, P).trace())
                logger.info(ks, "spin %s\n%s\n%s", s, lab_string, P)
            logger.info(ks, "-" * 79)

    if E_U.real < 0.0 and all(np.asarray(ks.U_val) > 0):
        logger.warn(ks, "E_U (%s) is negative...", E_U.real)
    vxc = lib.tag_array(vxc, E_U=E_U.real)
    return vxc

def energy_elec(mf, dm=None, h1e=None, vhf=None):
    """
    Electronic energy for UKSpU.
    """
    if h1e is None: h1e = mf.get_hcore()
    if dm is None: dm = mf.make_rdm1()
    if vhf is None or getattr(vhf, 'ecoul', None) is None:
        vhf = mf.get_veff(mf.mol, dm)

    e1 = (np.einsum('ij,ji->', h1e, dm[0]) +
          np.einsum('ij,ji->', h1e, dm[1]))
    tot_e = e1 + vhf.ecoul + vhf.exc + vhf.E_U
    mf.scf_summary['e1'] = e1.real
    mf.scf_summary['coul'] = vhf.ecoul.real
    mf.scf_summary['exc'] = vhf.exc.real
    mf.scf_summary['E_U'] = vhf.E_U.real

    logger.debug(mf, 'E1 = %s  Ecoul = %s  Exc = %s  EU = %s',
                 e1, vhf.ecoul, vhf.exc, vhf.E_U)
    return tot_e.real, vhf.ecoul + vhf.exc + vhf.E_U

class UKSpU(uks.UKS):
    """
    UKSpU class adapted for PBCs with k-point sampling.
    """

    _keys = {"U_idx", "U_val", "C_ao_lo", "U_lab", 'alpha'}

    get_veff = get_veff
    energy_elec = energy_elec
    to_hf = lib.invalid_method('to_hf')

    def __init__(self, mol, xc='LDA,VWN',
                 U_idx=[], U_val=[], C_ao_lo='minao', minao_ref='MINAO'):
        """
        DFT+U args:
            U_idx: can be
                   list of list: each sublist is a set of LO indices to add U.
                   list of string: each string is one kind of LO orbitals,
                                   e.g. ['Ni 3d', '1 O 2pz'], in this case,
                                   LO should be aranged as ao_labels order.
                   or a combination of these two.
            U_val: a list of effective U [in eV], i.e. U-J in Dudarev's DFT+U.
                   each U corresponds to one kind of LO orbitals, should have
                   the same length as U_idx.
            C_ao_lo: LO coefficients, can be
                     np.array, shape ((spin,), nao, nlo),
                     string, in 'minao'.
            minao_ref: reference for minao orbitals, default is 'MINAO'.

        Attributes:
            U_idx: same as the input.
            U_val: effectiv U-J [in AU]
            C_ao_loc: np.array
            alpha: the perturbation [in AU] used to compute U in LR-cDFT.
                Refs: Cococcioni and de Gironcoli, PRB 71, 035105 (2005)
        """
        super(self.__class__, self).__init__(mol, xc=xc)

        set_U(self, U_idx, U_val)

        if isinstance(C_ao_lo, str):
            if C_ao_lo.upper() == 'MINAO':
                self.C_ao_lo = make_minao_lo(self, minao_ref)
            else:
                raise NotImplementedError
        else:
            self.C_ao_lo = np.asarray(C_ao_lo)
        assert self.C_ao_lo.ndim == 2
        self.alpha = None

    def dump_flags(self, verbose=None):
        super().dump_flags(verbose)
        log = logger.new_logger(self, verbose)
        if log.verbose >= logger.INFO:
            from pyscf.pbc.dft.krkspu import _print_U_info
            _print_U_info(self, log)
        return self

    def nuc_grad_method(self):
        raise NotImplementedError

def linear_response_u(mf_plus_u, alphalist=(0.02, 0.05, 0.08)):
    '''
    Refs:
        [1] M. Cococcioni and S. de Gironcoli, Phys. Rev. B 71, 035105 (2005)
        [2] H. J. Kulik, M. Cococcioni, D. A. Scherlis, and N. Marzari, Phys. Rev. Lett. 97, 103001 (2006)
        [3] Heather J. Kulik, J. Chem. Phys. 142, 240901 (2015)
        [4] https://hjkgrp.mit.edu/tutorials/2011-05-31-calculating-hubbard-u/
        [5] https://hjkgrp.mit.edu/tutorials/2011-06-28-hubbard-u-multiple-sites/

    Args:
        alphalist :
            alpha parameters (in eV) are the displacements for the linear
            response calculations. For each alpha in this list, the DFT+U with
            U=u0+alpha, U=u0-alpha are evaluated. u0 is the U value from the
            reference mf_plus_u object, which will be treated as a standard DFT
            functional.
    '''
    assert isinstance(mf_plus_u, UKSpU)
    assert len(mf_plus_u.U_idx) > 0
    if not mf_plus_u.converged:
        mf_plus_u.run()
    assert mf_plus_u.converged
    # The bare density matrix without adding U
    bare_dm = mf_plus_u.make_rdm1()

    mf = mf_plus_u.copy()
    log = logger.new_logger(mf)

    alphalist = np.asarray(alphalist)
    alphalist = np.append(-alphalist[::-1], alphalist)

    C_ao_lo = mf.C_ao_lo
    ovlp = mf.get_ovlp()
    C_inv = [C_ao_lo[:,local_idx].conj().T.dot(ovlp) for local_idx in mf.U_idx]

    bare_occupancies = []
    final_occupancies = []
    for alpha in alphalist:
        mf.alpha = alpha / HARTREE2EV
        mf.kernel(dm0=bare_dm)
        local_occ = 0
        for c in C_inv:
            C_on_site = [c.dot(mf.mo_coeff[0]), c.dot(mf.mo_coeff[1])]
            rdm1_lo = mf.make_rdm1(C_on_site, mf.mo_occ)
            local_occ += rdm1_lo[0].trace() + rdm1_lo[1].trace()
        final_occupancies.append(local_occ)

        # The first iteration of SCF
        fock = mf.get_fock(dm=bare_dm)
        e, mo = mf.eig(fock, ovlp)
        local_occ = 0
        for c in C_inv:
            C_on_site = [c.dot(mo[0]), c.dot(mo[1])]
            rdm1_lo = mf.make_rdm1(C_on_site, mf.mo_occ)
            local_occ += rdm1_lo[0].trace() + rdm1_lo[1].trace()
        bare_occupancies.append(local_occ)
        log.info('alpha=%f bare_occ=%g final_occ=%g',
                 alpha, bare_occupancies[-1], final_occupancies[-1])

    chi0, occ0 = np.polyfit(alphalist, bare_occupancies, deg=1)
    chif, occf = np.polyfit(alphalist, final_occupancies, deg=1)
    log.info('Line fitting chi0 = %f x + %f', chi0, occ0)
    log.info('Line fitting chif = %f x + %f', chif, occf)
    Uresp = 1./chi0 - 1./chif
    log.note('Uresp = %f, chi0 = %f, chif = %f', Uresp, chi0, chif)
    return Uresp



#!/usr/bin/env python
# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Different FCI solvers are implemented to support different type of symmetry.
                    Symmetry
File                Point group   Spin singlet   Real hermitian*    Alpha/beta degeneracy
direct_spin0_symm   Yes           Yes            Yes                Yes
direct_spin1_symm   Yes           No             Yes                Yes
direct_spin0        No            Yes            Yes                Yes
direct_spin1        No            No             Yes                Yes
direct_uhf          No            No             Yes                No
direct_nosym        No            No             No**               Yes
fci_dhf_slow        No            No             No***              No

*  Real hermitian Hamiltonian implies (ij|kl) = (ji|kl) = (ij|lk) = (ji|lk)
** Hamiltonian is real but not hermitian, (ij|kl) != (ji|kl) ...
*** Hamiltonian is complex hermitian (DHF case) or real hermitian (GHF case)
'''

from pyscf.fci import cistring
from pyscf.fci import direct_spin0
from pyscf.fci import direct_spin1
from pyscf.fci import direct_uhf
from pyscf.fci import direct_spin0_symm
from pyscf.fci import direct_spin1_symm
from pyscf.fci import fci_dhf_slow
from pyscf.fci import addons
from pyscf.fci import rdm
from pyscf.fci import spin_op
from pyscf.fci.cistring import num_strings
from pyscf.fci.rdm import reorder_rdm
from pyscf.fci.spin_op import spin_square
from pyscf.fci.direct_spin1 import make_pspace_precond, make_diag_precond, FCIvector
from pyscf.fci import direct_nosym
from pyscf.fci import selected_ci
select_ci = selected_ci  # for backward compatibility
from pyscf.fci import selected_ci_spin0
from pyscf.fci import selected_ci_symm
from pyscf.fci import selected_ci_spin0_symm
from pyscf.fci.selected_ci import SelectedCI, SCI, SCIvector

def solver(mol=None, singlet=False, symm=None):
    if mol and symm is None:
        symm = mol.symmetry
    if symm:
        if singlet:
            return direct_spin0_symm.FCISolver(mol)
        else:
            return direct_spin1_symm.FCISolver(mol)
    else:
        if singlet:
            # The code for singlet direct_spin0 sometimes gets error of 
            # "State not singlet x.xxxxxxe-06" due to numerical issues.
            # Calling direct_spin1 is slightly slower but more robust than
            # direct_spin0 especially when combining to energy penalty method
            # (:func:`fix_spin_`)
            return direct_spin0.FCISolver(mol)
        else:
            return direct_spin1.FCISolver(mol)

def FCI(mol_or_mf, mo=None, singlet=False):
    '''FCI solver

    Args:
        mol_or_mf :
            A Mole object or an SCF object

    Kwargs:
        mo :
            Molecular orbital coefficients
        singlet :
            Whether to enable spin symmetry for S=0 RHF-based FCI solver.

    Returns:
        A FCI object
    '''
    from functools import reduce
    import numpy
    from pyscf import scf
    from pyscf import symm
    from pyscf import ao2mo
    from pyscf import lib
    if isinstance(mol_or_mf, scf.hf.SCF):
        mf = mol_or_mf
        mol = mf.mol
        if mo is None:
            mo = mf.mo_coeff
        is_uhf = isinstance(mf, scf.uhf.UHF)
        is_ghf = isinstance(mf, scf.ghf.GHF)
        is_dhf = isinstance(mf, scf.dhf.DHF)
    else:
        mf = None
        mol = mol_or_mf
        is_rhf = (mo is None or (isinstance(mo, numpy.ndarray) and mo.ndim == 2 and mo.shape[0] == mol.nao))
        is_ghf = (mo is not None and (isinstance(mo, numpy.ndarray) and mo.ndim == 2 and mo.shape[0] == 2 * mol.nao))
        is_dhf = (mo is not None and (isinstance(mo, numpy.ndarray) and mo.ndim == 2 and mo.shape[0] == 2 * mol.nao_2c()))
        is_uhf = not (is_rhf or is_ghf or is_dhf)

    if is_uhf:
        fcisolver = direct_uhf.FCI(mol)
    elif is_ghf or is_dhf:
        fcisolver = fci_dhf_slow.FCI(mol)
    else:
        fcisolver = solver(mol, singlet=(singlet and mol.spin==0))

    # Just create the FCI solver without initializing Hamiltonian
    if mo is None:
        return fcisolver

    nelec = getattr(mf, 'nelec', mol.nelec)

    if mf is None:
        hcore = scf.hf.get_hcore(mol)
        ecore = mol.energy_nuc()
    else:
        hcore = mf.get_hcore()
        ecore = mf.energy_nuc()

    if mf is None or mf._eri is None:
        if getattr(mol, 'pbc_intor', None):  # cell object has pbc_intor method
            raise NotImplementedError('Integral transformation for PBC object')
        eri_ao = mol
    else:
        eri_ao = mf._eri

    if mol.symmetry:
        if mf is None:
            s = mol.intor('int1e_ovlp')
        else:
            s = mf.get_ovlp()

        if is_uhf:
            orbsym = scf.uhf_symm.get_orbsym(mol, mo, s)
        elif is_ghf:
            orbsym = scf.ghf_symm.get_orbsym(mol, mo, s)
        elif is_dhf:
            orbsym = None
        else:
            orbsym = scf.hf_symm.get_orbsym(mol, mo, s)
    else:
        orbsym = None

    if is_uhf:
        h1e = [reduce(numpy.dot, (mo[0].conj().T, hcore, mo[0])),
               reduce(numpy.dot, (mo[1].conj().T, hcore, mo[1]))]
        eri_aa = ao2mo.kernel(eri_ao, (mo[0], mo[0], mo[0], mo[0]))
        eri_ab = ao2mo.kernel(eri_ao, (mo[0], mo[0], mo[1], mo[1]))
        eri_bb = ao2mo.kernel(eri_ao, (mo[1], mo[1], mo[1], mo[1]))
        eri = [eri_aa, eri_ab, eri_bb]
        norb = mo[0].shape[1]
    elif is_ghf:
        norb = mo.shape[1]
        h1e = reduce(numpy.dot, (mo.conj().T, hcore, mo))
        mo_a, mo_b = mo[:mol.nao], mo[mol.nao:]
        eri = ao2mo.restore(4, ao2mo.general(eri_ao, (mo_a, mo_a, mo_b, mo_b)), norb)
        eri = eri + eri.transpose(1, 0)
        eri += ao2mo.restore(4, ao2mo.full(eri_ao, mo_a), norb)
        eri += ao2mo.restore(4, ao2mo.full(eri_ao, mo_b), norb)
        eri = ao2mo.restore(1, eri, norb)
        nelec = sum(nelec)
    elif is_dhf:
        ecore = ecore.real
        ncore = 0
        ncas = mo.shape[1] // 4 - ncore
        nneg = mo.shape[1] // 4
        ncore += nneg
        mo_core = mo[:, nneg * 2:ncore * 2]
        mo_cas = mo[:, ncore * 2:ncore * 2 + ncas * 2]
        core_dm = mo_core @ mo_core.T.conj()
        if mf is not None:
            vj, vk = mf.get_jk(mol, core_dm)
        else:
            vj, vk = scf.dhf.get_jk_coulomb(mol, core_dm)
        hveff = vj - vk
        ecore += numpy.sum(core_dm.T * (hcore + 0.5 * hveff)).real
        c1 = 0.5 / lib.param.LIGHT_SPEED
        h1e = reduce(numpy.dot, (mo_cas.conj().T, hcore + hveff, mo_cas))
        mo_l, mo_s = mo_cas[:mol.nao_2c()], mo_cas[mol.nao_2c():]
        eri = ao2mo.general(mol, (mo_l, mo_l, mo_s, mo_s), intor="int2e_spsp2_spinor", aosym=4)
        eri = (eri + eri.transpose(1, 0)) * c1 ** 2
        eri += ao2mo.full(mol, mo_l, intor="int2e_spinor", aosym=4)
        eri += ao2mo.full(mol, mo_s, intor="int2e_spsp1spsp2_spinor", aosym=4) * c1 ** 4
        if mf is not None and mf.with_gaunt:
            p = "int2e_breit_" if mf.with_breit else "int2e_"
            eri_lsls = ao2mo.general(mol, (mo_l, mo_s, mo_l, mo_s), intor=p + "ssp1ssp2_spinor", aosym=1, comp=1)
            eri_slsl = eri_lsls.reshape((ncas * 2,) * 4).transpose(3, 2, 1, 0).conj().reshape((ncas * ncas * 4,) * 2)
            eri_lssl = ao2mo.general(mol, (mo_l, mo_s, mo_s, mo_l), intor=p + "ssp1sps2_spinor", aosym=1, comp=1)
            eri_slls = eri_lssl.transpose(1, 0)
            if mf.with_breit:
                eri += (eri_lsls + eri_slsl + eri_lssl + eri_slls) * c1 ** 2
            else:
                eri -= (eri_lsls + eri_slsl + eri_lssl + eri_slls) * c1 ** 2
        eri = eri.reshape((ncas * 2,) * 4)
        norb = ncas * 2
        nelec = sum(nelec)
    else:
        h1e = reduce(numpy.dot, (mo.conj().T, hcore, mo))
        eri = ao2mo.kernel(eri_ao, mo)
        norb = mo.shape[1]

    fcisolver_class = fcisolver.__class__
    class CISolver(fcisolver_class):
        def __init__(self, mol=None):
            fcisolver_class.__init__(self, mol)
            self.orbsym = orbsym

        def kernel(self, h1e=h1e, eri=eri, norb=norb, nelec=nelec, ci0=None,
                   ecore=ecore, **kwargs):
            return fcisolver_class.kernel(self, h1e, eri, norb, nelec, ci0,
                                          ecore=ecore, **kwargs)
    cisolver = CISolver(mol)
    cisolver.__dict__.update(fcisolver.__dict__)
    cisolver.orbsym = orbsym
    return cisolver



#!/usr/bin/env python
# Copyright 2018-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Peng Bao <baopeng@iccas.ac.cn>
#         Qiming Sun <osirpt.sun@gmail.com>
#

'''
semi-grid Coulomb and eXchange without differential density matrix

To lower the scaling of coulomb and exchange matrix construction for large system, one
coordinate is analytical and the other is grid. The traditional two electron
integrals turn to analytical one electron integrals and numerical integration
based on grid.(see Friesner, R. A. Chem. Phys. Lett. 1985, 116, 39)

Minimizing numerical errors using overlap fitting correction.(see
Lzsak, R. et. al. J. Chem. Phys. 2011, 135, 144105)
Grid screening for weighted AO value and DktXkg.
Two SCF steps: coarse grid then fine grid. There are 5 parameters can be changed:
# threshold for Xg and Fg screening
gthrd = 1e-10
# initial and final grids level
grdlvl_i = 0
grdlvl_f = 1
# norm_ddm threshold for grids change
thrd_nddm = 0.03
# set block size to adapt memory
sblk = 200

Set mf.direct_scf = False because no traditional 2e integrals
'''


import ctypes
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.df.incore import aux_e2
from pyscf.gto import moleintor
from pyscf.scf import _vhf
from pyscf.dft import gen_grid


def get_jk_favork(sgx, dm, hermi=1, with_j=True, with_k=True,
                  direct_scf_tol=1e-13):
    t0 = logger.process_clock(), logger.perf_counter()
    mol = sgx.mol
    grids = sgx.grids
    gthrd = sgx.grids_thrd

    dms = numpy.asarray(dm)
    dm_shape = dms.shape
    nao = dm_shape[-1]
    dms = dms.reshape(-1,nao,nao)
    nset = dms.shape[0]

    if sgx.debug:
        batch_nuc = _gen_batch_nuc(mol)
    else:
        batch_jk = _gen_jk_direct(mol, 's2', with_j, with_k, direct_scf_tol,
                                  sgx._opt, sgx.pjs)
    t1 = logger.timer_debug1(mol, "sgX initialization", *t0)

    sn = numpy.zeros((nao,nao))
    vj = numpy.zeros_like(dms)
    vk = numpy.zeros_like(dms)

    ngrids = grids.coords.shape[0]
    max_memory = sgx.max_memory - lib.current_memory()[0]
    sblk = sgx.blockdim
    blksize = min(ngrids, max(4, int(min(sblk, max_memory*1e6/8/nao**2))))
    tnuc = 0, 0
    for i0, i1 in lib.prange(0, ngrids, blksize):
        coords = grids.coords[i0:i1]
        weights = grids.weights[i0:i1,None]
        ao = mol.eval_gto('GTOval', coords)
        wao = ao * grids.weights[i0:i1,None]
        sn += lib.dot(ao.T, wao)

        fg = lib.einsum('gi,xij->xgj', wao, dms)
        mask = numpy.zeros(i1-i0, dtype=bool)
        for i in range(nset):
            mask |= numpy.any(fg[i]>gthrd, axis=1)
            mask |= numpy.any(fg[i]<-gthrd, axis=1)
        if not numpy.all(mask):
            ao = ao[mask]
            wao = wao[mask]
            fg = fg[:,mask]
            coords = coords[mask]
            weights = weights[mask]

        if sgx.debug:
            tnuc = tnuc[0] - logger.process_clock(), tnuc[1] - logger.perf_counter()
            gbn = batch_nuc(mol, coords)
            tnuc = tnuc[0] + logger.process_clock(), tnuc[1] + logger.perf_counter()
            if with_j:
                jg = numpy.einsum('gij,xij->xg', gbn, dms)
            if with_k:
                gv = lib.einsum('gvt,xgt->xgv', gbn, fg)
            gbn = None
        else:
            tnuc = tnuc[0] - logger.process_clock(), tnuc[1] - logger.perf_counter()
            jg, gv = batch_jk(mol, coords, dms, fg.copy(), weights)
            tnuc = tnuc[0] + logger.process_clock(), tnuc[1] + logger.perf_counter()

        if with_j:
            xj = lib.einsum('gv,xg->xgv', ao, jg)
            for i in range(nset):
                vj[i] += lib.einsum('gu,gv->uv', wao, xj[i])
        if with_k:
            for i in range(nset):
                vk[i] += lib.einsum('gu,gv->uv', ao, gv[i])
        jg = gv = None

    t2 = logger.timer_debug1(mol, "sgX J/K builder", *t1)
    tdot = t2[0] - t1[0] - tnuc[0] , t2[1] - t1[1] - tnuc[1]
    logger.debug1(sgx, '(CPU, wall) time for integrals (%.2f, %.2f); '
                  'for tensor contraction (%.2f, %.2f)',
                  tnuc[0], tnuc[1], tdot[0], tdot[1])

    ovlp = mol.intor_symmetric('int1e_ovlp')
    proj = scipy.linalg.solve(sn, ovlp)

    if with_j:
        vj = lib.einsum('pi,xpj->xij', proj, vj)
        vj = (vj + vj.transpose(0,2,1))*.5
    if with_k:
        vk = lib.einsum('pi,xpj->xij', proj, vk)
        if hermi == 1:
            vk = (vk + vk.transpose(0,2,1))*.5
    logger.timer(mol, "vj and vk", *t0)
    return vj.reshape(dm_shape), vk.reshape(dm_shape)


def get_jk_favorj(sgx, dm, hermi=1, with_j=True, with_k=True,
                  direct_scf_tol=1e-13):
    t0 = logger.process_clock(), logger.perf_counter()
    mol = sgx.mol
    grids = sgx.grids
    gthrd = sgx.grids_thrd

    dms = numpy.asarray(dm)
    dm_shape = dms.shape
    nao = dm_shape[-1]
    dms = dms.reshape(-1,nao,nao)
    nset = dms.shape[0]

    if sgx.debug:
        batch_nuc = _gen_batch_nuc(mol)
    else:
        batch_jk = _gen_jk_direct(mol, 's2', with_j, with_k, direct_scf_tol,
                                  sgx._opt, sgx.pjs)

    sn = numpy.zeros((nao,nao))
    ngrids = grids.coords.shape[0]
    max_memory = sgx.max_memory - lib.current_memory()[0]
    sblk = sgx.blockdim
    blksize = min(ngrids, max(4, int(min(sblk, max_memory*1e6/8/nao**2))))
    for i0, i1 in lib.prange(0, ngrids, blksize):
        coords = grids.coords[i0:i1]
        ao = mol.eval_gto('GTOval', coords)
        wao = ao * grids.weights[i0:i1,None]
        sn += lib.dot(ao.T, wao)

    ovlp = mol.intor_symmetric('int1e_ovlp')
    proj = scipy.linalg.solve(sn, ovlp)
    proj_dm = lib.einsum('ki,xij->xkj', proj, dms)

    t1 = logger.timer_debug1(mol, "sgX initialization", *t0)
    vj = numpy.zeros_like(dms)
    vk = numpy.zeros_like(dms)
    tnuc = 0, 0
    for i0, i1 in lib.prange(0, ngrids, blksize):
        coords = grids.coords[i0:i1]
        weights = grids.weights[i0:i1,None]
        ao = mol.eval_gto('GTOval', coords)
        wao = ao * grids.weights[i0:i1,None]

        fg = lib.einsum('gi,xij->xgj', wao, proj_dm)
        mask = numpy.zeros(i1-i0, dtype=bool)
        for i in range(nset):
            mask |= numpy.any(fg[i]>gthrd, axis=1)
            mask |= numpy.any(fg[i]<-gthrd, axis=1)
        if not numpy.all(mask):
            ao = ao[mask]
            fg = fg[:,mask]
            coords = coords[mask]
            weights = weights[mask]

        if with_j:
            rhog = numpy.einsum('xgu,gu->xg', fg, ao)
        else:
            rhog = None

        if sgx.debug:
            tnuc = tnuc[0] - logger.process_clock(), tnuc[1] - logger.perf_counter()
            gbn = batch_nuc(mol, coords)
            tnuc = tnuc[0] + logger.process_clock(), tnuc[1] + logger.perf_counter()
            if with_j:
                jpart = numpy.einsum('guv,xg->xuv', gbn, rhog)
            if with_k:
                gv = lib.einsum('gtv,xgt->xgv', gbn, fg)
            gbn = None
        else:
            tnuc = tnuc[0] - logger.process_clock(), tnuc[1] - logger.perf_counter()
            if with_j: rhog = rhog.copy()
            jpart, gv = batch_jk(mol, coords, rhog, fg.copy(), weights)
            tnuc = tnuc[0] + logger.process_clock(), tnuc[1] + logger.perf_counter()

        if with_j:
            vj += jpart
        if with_k:
            for i in range(nset):
                vk[i] += lib.einsum('gu,gv->uv', ao, gv[i])
        jpart = gv = None

    t2 = logger.timer_debug1(mol, "sgX J/K builder", *t1)
    tdot = t2[0] - t1[0] - tnuc[0] , t2[1] - t1[1] - tnuc[1]
    logger.debug1(sgx, '(CPU, wall) time for integrals (%.2f, %.2f); '
                  'for tensor contraction (%.2f, %.2f)',
                  tnuc[0], tnuc[1], tdot[0], tdot[1])

    for i in range(nset):
        lib.hermi_triu(vj[i], inplace=True)
    if with_k and hermi == 1:
        vk = (vk + vk.transpose(0,2,1))*.5
    logger.timer(mol, "vj and vk", *t0)
    return vj.reshape(dm_shape), vk.reshape(dm_shape)

def _gen_batch_nuc(mol):
    '''Coulomb integrals of the given points and orbital pairs'''
    cintopt = gto.moleintor.make_cintopt(mol._atm, mol._bas, mol._env, 'int3c2e')
    def batch_nuc(mol, grid_coords, out=None):
        fakemol = gto.fakemol_for_charges(grid_coords)
        j3c = aux_e2(mol, fakemol, intor='int3c2e', aosym='s2ij', cintopt=cintopt)
        return lib.unpack_tril(j3c.T, out=out)
    return batch_nuc

def _gen_jk_direct(mol, aosym, with_j, with_k, direct_scf_tol, sgxopt=None, pjs=False):
    '''Contraction between sgX Coulomb integrals and density matrices

    J: einsum('guv,xg->xuv', gbn, dms) if dms == rho at grid,
    or einsum('gij,xij->xg', gbn, dms) if dms are density matrices

    K: einsum('gtv,xgt->xgv', gbn, fg)
    '''
    if sgxopt is None:
        from pyscf.sgx import sgx
        sgxopt = sgx._make_opt(mol, pjs, direct_scf_tol)
    sgxopt.direct_scf_tol = direct_scf_tol

    ncomp = 1
    nao = mol.nao
    cintor = _vhf._fpointer(sgxopt._intor)
    fdot = _vhf._fpointer('SGXdot_nrk')
    drv = _vhf.libcvhf.SGXnr_direct_drv

    def jk_part(mol, grid_coords, dms, fg, weights):
        atm, bas, env = mol._atm, mol._bas, mol._env
        ngrids = grid_coords.shape[0]
        env = numpy.append(env, grid_coords.ravel())
        env[gto.NGRIDS] = ngrids
        env[gto.PTR_GRIDS] = mol._env.size
        if pjs:
            sgxopt.set_dm(fg / numpy.sqrt(numpy.abs(weights[None,:])),
                          mol._atm, mol._bas, env)

        ao_loc = moleintor.make_loc(bas, sgxopt._intor)
        shls_slice = (0, mol.nbas, 0, mol.nbas)

        fg = numpy.ascontiguousarray(fg.transpose(0,2,1))

        vj = vk = None
        fjk = []
        dmsptr = []
        vjkptr = []
        if with_j:
            if dms[0].ndim == 1:  # the value of density at each grid
                vj = numpy.zeros((len(dms),ncomp,nao,nao))[:,0]
                for i, dm in enumerate(dms):
                    dmsptr.append(dm.ctypes.data_as(ctypes.c_void_p))
                    vjkptr.append(vj[i].ctypes.data_as(ctypes.c_void_p))
                    fjk.append(_vhf._fpointer('SGXnr'+aosym+'_ijg_g_ij'))
            else:
                vj = numpy.zeros((len(dms),ncomp,ngrids))[:,0]
                for i, dm in enumerate(dms):
                    dmsptr.append(dm.ctypes.data_as(ctypes.c_void_p))
                    vjkptr.append(vj[i].ctypes.data_as(ctypes.c_void_p))
                    fjk.append(_vhf._fpointer('SGXnr'+aosym+'_ijg_ji_g'))
        if with_k:
            vk = numpy.zeros((len(fg),ncomp,nao,ngrids))[:,0]
            for i, dm in enumerate(fg):
                dmsptr.append(dm.ctypes.data_as(ctypes.c_void_p))
                vjkptr.append(vk[i].ctypes.data_as(ctypes.c_void_p))
                fjk.append(_vhf._fpointer('SGXnr'+aosym+'_ijg_gj_gi'))

        n_dm = len(fjk)
        fjk = (ctypes.c_void_p*(n_dm))(*fjk)
        dmsptr = (ctypes.c_void_p*(n_dm))(*dmsptr)
        vjkptr = (ctypes.c_void_p*(n_dm))(*vjkptr)

        drv(cintor, fdot, fjk, dmsptr, vjkptr, n_dm, ncomp,
            (ctypes.c_int*4)(*shls_slice),
            ao_loc.ctypes.data_as(ctypes.c_void_p),
            sgxopt._cintopt, ctypes.byref(sgxopt._this),
            atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(mol.natm),
            bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(mol.nbas),
            env.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(env.shape[0]),
            ctypes.c_int(2 if aosym == 's2' else 1))
        if vk is not None:
            vk = vk.transpose(0,2,1)
            vk = numpy.ascontiguousarray(vk)
        return vj, vk
    return jk_part


# pre for get_k
# Use default mesh grids and weights
def get_gridss(mol, level=1, gthrd=1e-10):
    Ktime = (logger.process_clock(), logger.perf_counter())
    grids = gen_grid.Grids(mol)
    grids.level = level
    grids.build()

    ngrids = grids.weights.size
    mask = []
    for p0, p1 in lib.prange(0, ngrids, 10000):
        ao_v = mol.eval_gto('GTOval', grids.coords[p0:p1])
        ao_v *= grids.weights[p0:p1,None]
        wao_v0 = ao_v
        mask.append(numpy.any(wao_v0>gthrd, axis=1) |
                    numpy.any(wao_v0<-gthrd, axis=1))

    mask = numpy.hstack(mask)
    grids.coords = grids.coords[mask]
    grids.weights = grids.weights[mask]
    logger.debug(mol, 'threshold for grids screening %g', gthrd)
    logger.debug(mol, 'number of grids %d', grids.weights.size)
    logger.timer_debug1(mol, "Xg screening", *Ktime)
    return grids

get_jk = get_jk_favorj
